from .func import TanHDist, SoftDist, HardDist, DoubleSoftDist, CubicDist, RectifierDist, ArcTanDist, ExponentialDist
from .bitcrusher import BitCrusher

__all__ = [
    "TanHDist",
    "SoftDist",
    "HardDist",
    "DoubleSoftDist",
    "CubicDist",
    "RectifierDist",
    "BitCrusher",
    "ExponentialDist",
    "ArcTanDist"
]


